<?php
namespace SupplyTek\Services;

use Doctrine\ORM\EntityManager;
use SupplyTek\Utilities\SessionHandler;
use SupplyTek\Libraries\NewCode;

/**
 * SupplyTek Approval/Disapproval Model
 * 
 * Connects to the Database to handle queries 
 * related to the approval/dusapproval workflow. 
 *  
 * @author Rômulo Alves Lousada
 */
class ApprovalWorkflowService
{
    protected $connection;
    protected $session;
    protected $newCode;

    /**
     * Class constructor
     * 
     * Initialize an object's properties upon creation of the object.
     * 
     * @param EntityManager $entityManager Object to manage database connection and queries
     * @param SessionHandler $sessionHandler Object that allows to manipulate PHP Session
     * @param NewCodeSC $newCodeSC Object to generate new code without commiting
     * 
     * @return void
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function __construct(EntityManager $entityManager, SessionHandler $sessionHandler, NewCode $newCodeSC)
    {
        $this->connection = $entityManager->getConnection();
        $this->session = $sessionHandler;
        $this->newCode = $newCodeSC;
    }
    
    /**************************************************************************/
    /*                                UTILITIES                               */
    /**************************************************************************/
    
    /**
     * Generate New Code without Commit
     * 
     * Generates and retrieves the new code to the table sent, needed to save
     * as a PK from the table.
     * 
     * @param string $tableName String containing the table name to generate the code
     *
     * @return int
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function generateCodeSC($tableName)
    {
        return (int) $this->newCode->getCodeByTableName($tableName, $this->session->getSessionData("NRORG"));
    }
    
    /**************************************************************************/
    /*                             WORKFLOW START                             */
    /**************************************************************************/
    
    /**
     * Returns the authority number
     * 
     * @param array $parameters Array with request data received to make the process
     *
     * @return mixed
     * @throws Exception
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function getAuthorityNumber($parameters)
    {
        try {
            $sql = "SELECT 
                        NRCONFIGALCADA
                    FROM 
                        ALCADAGERFILI
                    WHERE 
                        CDFILIAL = :P_CDFILDEST
                        AND CDOPERADOR = :P_CDOPERADOR";
                    
            $params = array(
                "P_CDFILDEST" => $parameters["P_CDFILDEST"],
                "P_CDOPERADOR" => $parameters["P_CDOPERADOR"]
            );
    
            $std = $this->connection->prepare($sql);
            $std->execute($params);
            
            $result = $std->fetchAll();
            
            return isset($result[0]) ? $result[0]["NRCONFIGALCADA"] : false;
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    /**
     * Save registry of approval on APROVALCADA
     * 
     * @param array $parameters Array with request data received to make the process
     *
     * @return int
     * @throws Exception
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function saveApprovalRegistry($parameters)
    {
        try {
            $nrSeq = $this->generateCodeSC("APROVALCADA");
            
            $sql = "INSERT INTO APROVALCADA 
                    ( 
                        NRAPROVALCADA, 
                        NRCONFIGALCADA, 
                        CDFILDEST, 
                        CDFILLANC, 
                        NRDEMANDA, 
                        CDOPERADOR, 
                        IDTPDEMANDA, 
                        IDCLASSIFIC, 
                        IDOPERACAOALC, 
                        IDAPROVADO, 
                        DTINCLUSAO 
                    ) 
                    VALUES 
                    ( 
                        :P_NRAPROVALCADA, 
                        :P_NRCONFIGALCADA, 
                        :P_CDFILDEST, 
                        :P_CDFILLANC, 
                        :P_NRDEMANDA, 
                        :P_CDOPERADOR, 
                        'AF', 
                        'PADRAO', 
                        'LANCAMENTO', 
                        'PENDENTE', 
                        SYSDATE 
                    )";
            
            $params = array(
                "P_NRAPROVALCADA" => $nrSeq,
                "P_NRCONFIGALCADA" => $parameters["P_NRCONFIGALCADA"],
                "P_CDFILDEST" => $parameters["P_CDFILDEST"],
                "P_CDFILLANC" => $parameters["P_CDFILLANC"],
                "P_NRDEMANDA" => $parameters["P_NRDEMANDA"],
                "P_CDOPERADOR" => $parameters["P_CDOPERADOR"]
            );
    
            $std = $this->connection->prepare($sql);
            $std->execute($params);
            
            return $nrSeq;
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    /**
     * Update AUTOFORNE with Authority data.
     * 
     * @param array $parameters Array with request data received to make the process
     *
     * @return void
     * @throws Exception
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function updateAFWithAuthorityData($parameters)
    {
        try {
            $sql = "UPDATE 
                        AUTOFORNE
                    SET 
                        IDBAIXAAF = 6, 
                        NRCONFIGALCADA = :P_NRCONFIGALCADA, 
                        IDSITAFALCADA = 'PENDENTE'
                    WHERE 
                        CDFILDEST = :P_CDFILDEST
                        AND CDFILLANC = :P_CDFILLANC
                        AND NRAUTOFORN = :P_NRDEMANDA";
            
            $params = array(
                "P_NRCONFIGALCADA" => $parameters["P_NRCONFIGALCADA"],
                "P_CDFILDEST" => $parameters["P_CDFILDEST"],
                "P_CDFILLANC" => $parameters["P_CDFILLANC"],
                "P_NRDEMANDA" => $parameters["P_NRDEMANDA"]
            );
    
            $std = $this->connection->prepare($sql);
            $std->execute($params);
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    /**
     * Returns user data needed on email.
     * 
     * @param array $parameters Array with request data received to make the process
     *
     * @return mixed
     * @throws Exception
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function getUserData($parameters)
    {
        try {
            $sql = "SELECT 
                        NMOPERADOR,
                        DSEMAILOPER
                    FROM 
                        OPERADOR
                    WHERE 
                        CDOPERADOR = :P_CDOPERADOR";
                    
            $params = array(
                "P_CDOPERADOR" => $parameters["P_CDOPERADOR"]
            );
    
            $std = $this->connection->prepare($sql);
            $std->execute($params);
            
            $result = $std->fetchAll();
            
            return isset($result[0]) ? $result[0] : false;
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    /**
     * Return required data to send next approver email
     * 
     * Searches the database and retrieves User data needed
     * on the email workflow 
     * 
     * @param array $parameters Array with request data received to make the process
     *
     * @return mixed
     * @throws Exception
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function getMailData($parameters)
    {
        try {
            $sql = "SELECT 
                        AUTOFORNE.CDFILDEST, 
                        FILIAL.NMFILIAL AS NMFILDEST, 
                        AUTOFORNE.DTAUTOFORN, 
                        (
                            SELECT 
                                MAX(DTPROGENTAF)
                            FROM 
                                ITEMAUTF
                            WHERE 
                                ITEMAUTF.CDFILDEST = AUTOFORNE.CDFILDEST 
                                AND ITEMAUTF.CDFILLANC = AUTOFORNE.CDFILLANC
                                AND ITEMAUTF.NRAUTOFORN = AUTOFORNE.NRAUTOFORN
                        ) AS DTPROGENTAF,
                        AUTOFORNE.VRTOTALAF, 
                        COMPRADOR.CDOPERADOR || ' - ' || COMP.NMOPERADOR AS COMPRADOR,
                        OPERADOR.NMOPERADOR AS NMSOLICITANTE, 
                        OPERADOR.DSEMAILOPER AS EMAILSOLICITANTE,
                        FORMATMASK(FORNECEDOR.IDTPIJURFORN, FORNECEDOR.NRINSJURFORN) AS NRINSJURFORN, 
                        FORNECEDOR.NMFANTFORN
                    FROM 
                        AUTOFORNE, 
                        OPERADOR, 
                        FILIAL, 
                        FORNECEDOR, 
                        OPERADOR COMP, 
                        COMPRADOR, 
                        PARAMGERALECOMMERCE,
                        TABLE
                        (
                            DADOS_FORN_EMAIL
                            (
                                FORNECEDOR.CDFORNECED, 'S', '', AUTOFORNE.NRORG 
                            )
                        ) EMAILFORNE
                    WHERE 
                        AUTOFORNE.CDFILDEST = :P_CDFILDEST 
                        AND AUTOFORNE.CDFILLANC = :P_CDFILLANC
                        AND AUTOFORNE.NRAUTOFORN = :P_NRAUTOFORN
                        AND AUTOFORNE.CDOPERLANCAF = OPERADOR.CDOPERADOR
                        AND AUTOFORNE.CDFILDEST = FILIAL.CDFILIAL
                        AND AUTOFORNE.CDFORNECED = FORNECEDOR.CDFORNECED
                        AND AUTOFORNE.CDOPERADOR = COMPRADOR.CDOPERADOR
                        AND COMPRADOR.CDOPERADOR = COMP.CDOPERADOR
                        AND FILIAL.CDFILCENTCMP = PARAMGERALECOMMERCE.CDFILIAL (+)";
            
            $params = array(
                "P_CDFILDEST" => $parameters["P_CDFILDEST"],
                "P_CDFILLANC" => $parameters["P_CDFILLANC"],
                "P_NRAUTOFORN" => $parameters["P_NRDEMANDA"]
            );
            
            $std = $this->connection->prepare($sql);
            $std->execute($params);
            
            $result = $std->fetchAll();
            
            return isset($result[0]) ? $result[0] : false;
        } catch(\Exception $e) {
            throw $e;
        }
    }
}