<?php
namespace SupplyTek\Utilities;

use SupplyTek\Utilities\JWT;
use Zeedhi\Framework\DTO\Request;
use SupplyTek\Utilities\SessionHandler;

/**
 * Checks if Token is valid for each request
 * 
 * For every request made from the frontend to the backend,
 * check the route, and if it's not the login route, validates
 * the Token. If the validation fail at any points, clear the session
 * variables and redirect the request to the login page.
 * 
 * @author Rômulo Alves Lousada
 */
class JWTChecker extends \Zeedhi\Framework\Events\PreDispatch\Listener 
{
    protected $session;

    /**
     * Class constructor
     * 
     * Initialize an object's properties upon creation of the object.
     * 
     * @param object $sessionHandler Object that allows to manipulate PHP Session
     * 
     * @return void
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function __construct ($sessionHandler) 
    {
        $this->session = $sessionHandler;
    }

    /**
     * Checks request route and Token sent
     * 
     * Listen to request, then throw an exception or
     * continue execution after validating the token
     * received.
     *
     * @param Request $request
     * 
     * @return void
     * @throws Exception [Based on what validation it failed to pass]
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function preDispatch(Request $request) 
    {
        $route = $request->getRoutePath();
        
        if (!$this->checkRouteAuthorized($route)) {
            $requestHeaders = apache_request_headers();
            $token = "";
                      
            if (isset($requestHeaders['Authorization'])) {
                $token = trim($requestHeaders['Authorization']);

                if (preg_match('/Bearer\s(\S+)/', $token, $matches)) {
                    $token = $matches[1];
                }
            }

            if ($token != "") {
                if(!JWT::validateExpiration($token)) {
                    $this->clearSessionData();
                    throw new \Exception('EXPIRED_TOKEN', 401);
                }

                if(!JWT::validateToken($token)){
                    $this->clearSessionData();
                    throw new \Exception('INVALID_TOKEN', 401);
                }
            } else {
                $this->clearSessionData();
                throw new \Exception('FORBIDDEN_ACCESS', 401);
            }
        }
    }
    
    /**
     * Check if the route is authorized.
     * 
     * Checks if sent route is inside the array of authorized
     * routes and returns the result.
     * 
     * @param string $route String with route name to be validated
     * 
     * @return bool
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function checkRouteAuthorized($route)
    {
        return in_array($route, $this->returnAuthorizedRoutes());
    }
    
    /**
     * Return authorized routes.
     * 
     * Retrieves an array with routes that doesn't need to validate
     * the JWT Token.
     * 
     * @return array
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function returnAuthorizedRoutes()
    {
        return array (
            "/auth", 
            "/send-email-new-password",
            "/send-new-password",
            "/metadata/login",
            "/metadata/modal-login",
            "/metadata/new-modal-login",
            "/metadata/modal-privacy-policy",
            "/find-privacy-policy-by-authentication"
        );
    }

    /**
     * Clears Session
     * 
     * Clear all PHP Session Data after any failure in
     * validating the Token.
     *
     * @return void
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function clearSessionData()
    {
        $this->session->clearSession();
    }
}